#include <cstdio>
#include <algorithm>
#include <functional>
#include <numeric>

#include "data.hpp"


int stem::byte_size() const
{
    int ret=1;
    for(std::list<param>::const_iterator it=params.begin();
        it!=params.end(); ++it){
        ret+=it->byte_size();
    }
    return ret;
}

int param::byte_size() const
{
    int ret=0;
    if(type==param::exp_type) return exp.size();
    else {
        for(std::list<stem>::const_iterator it=stats.begin();
            it!=stats.end(); ++it){
            ret+=it->byte_size();
        }
    }
    return ret;
}

int function::byte_size() const
{
    int ret=argument.size()+1;
    for(std::list<stem>::const_iterator it=stats.begin();
        it!=stats.end(); ++it){
        ret+=it->byte_size();
    }
    return ret;
}

int byte_size(std::list<function> const& pgm)
{
    int ret=0;
    for(std::list<function>::const_iterator it=pgm.begin(); it!=pgm.end();
        ++it){
            ret+=it->byte_size();
    }
    return ret-1>0?ret-1:0;
}

void print(std::FILE* fp, term const& t)
{
	if(t.type==term::number_type){
		fprintf(fp, "%c%d", t.neg?'-':'+', t.number);
	} else {
		fprintf(fp, "%c%c", t.neg?'-':'+', t.arg);
	}
}

void print(std::FILE* fp, exp const& e)
{
	if(e.empty()) return;
	exp::const_iterator it=e.begin();
	if(it->type==term::number_type) fprintf(fp, "%d", it->number*(it->neg?-1:1));
	else {
		if(it->neg) fprintf(fp, "-");
		fprintf(fp, "%c", it->arg);
	}
	++it;
	for(; it!=e.end(); ++it){
		print(fp,*it);
	}
}

void print(std::FILE* fp, std::list<stem> const& stats)
{
	for(std::list<stem>::const_iterator it=stats.begin(); it!=stats.end(); ++it){
		print(fp,*it);
	}
}

void print(std::FILE* fp, std::list<param> const& params)
{
	if(params.empty())return;

	std::list<param>::const_iterator en=params.end();
	en--;
	for(std::list<param>::const_iterator it=params.begin(); it!=en; ++it){
		print(fp,*it);fprintf(fp,",");
	}
	print(fp, params.back());
}

void print(std::FILE* fp, stem const& s)
{
	switch(s.type){
		case stem::s: fprintf(fp, "s"); break;
		case stem::l: fprintf(fp, "l"); break;
		case stem::r: fprintf(fp, "r"); break;
		case stem::arg_type: fprintf(fp, "%c", s.name); break;
		case stem::fun_call_type:
			fprintf(fp, "%c", s.name);
			if(! s.params.empty()){
				fprintf(fp, "(");
				print(fp, s.params);
				fprintf(fp, ")");
			}
			break;
	}
}

void print(std::FILE* fp, param const& p)
{
	if(p.type==param::exp_type){
		print(fp, p.exp);
	} else {
		print(fp, p.stats);
	}
}

void print(std::FILE* fp, function const& f)
{
	fprintf(fp, "%c", f.name);

	if(! f.argument.empty()){
		fprintf(fp,"(");
		std::list<char>::const_iterator end=f.argument.end();
		end--;
		for(std::list<char>::const_iterator it=f.argument.begin(); it!=end; ++it)
			fprintf(fp, "%c,", *it);
		fprintf(fp,"%c)",f.argument.back());
	}
	fprintf(fp, ":");
	print(fp, f.stats);
}
